/*
	$Id$

	Copyright (C) 2000, 2001 Syllabub
	Maroontress Fast Software.
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xresource.h>
#include <X11/Xlocale.h>
#include <X11/extensions/shape.h>
#include <X11/xpm.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Resource.h"
#include "WCString.h"
#include "Text.h"
#include "PreeditStringPainter.h"
#include "BorderDecorator.h"
#include "property.h"
#ifdef DEBUG
#include "detection.h"
#endif

#define THIS_CLASS "XIMReceiver"
#define THIS_NAME "imreceiver"

#define GET_STRING        0x01
#define GET_FEEDBACK      0x02
#define GET_CARET         0x04
#define GET_SPOT_LOCATION 0x08
#define GET_FONTSET       0x10

typedef struct IMReceiver {
    struct IMReceiver *next;

    Window target;
    Window window;
    PreeditStringPainter *psp;
    BorderDecorator *bd;

    int schedule;
    int mapcontrol;
} IMReceiver;

static Display *disp;
static IMReceiver *top = NULL;
static XContext Context;
static Atom RECEIVER;
static Atom PREEDIT_WINDOW;
static Atom PREEDIT_STRING;
static Atom PREEDIT_FEEDBACK;
static Atom PREEDIT_CARET;
static Atom PREEDIT_SPOT_LOCATION;
static Atom PREEDIT_FONTSET;
static Atom STATUS_STRING;
static Atom PREDICT_STRING;
static IMReceiver TheRootPreedit;
static IMReceiver TheRootStatus;
static IMReceiver TheRootPredict;
static int MainPower = True;
static int MapControl = False;

static SylSetting
    AutoHiddenMode = {"autoHidden", "AutoHidden", "False", NULL};

static void
WCSMapReceiver(IMReceiver *x, wchar_t *wcs)
{
    SetWCStringPreeditStringPainter(x->psp, wcs); 
    XMapRaised(disp, x->window);
}

static void
MBSMapReceiver(IMReceiver *x, char *mbs)
{
    SetMBStringPreeditStringPainter(x->psp, mbs); 
    XMapRaised(disp, x->window);
}

static void
UnmapReceiver(IMReceiver *x)
{
    wchar_t null = 0;

    SetWCStringPreeditStringPainter(x->psp, &null);
    if (x->mapcontrol == True) {
	XUnmapWindow(disp, x->window);
    }
}

static void
GetString(IMReceiver *x, Atom atom)
{
    XTextProperty text;
    wchar_t **list;
    int n_lists, n;

    if (XGetTextProperty(disp, x->target, &text, atom) == 0) {
	UnmapReceiver(x);
	return;
    } 
    if ((n = XwcTextPropertyToTextList(disp, &text, &list, &n_lists)) < 0) {
	XFree(text.value);
	return;
    }
    XFree(text.value);

    if (n_lists > 0) {		
	WCSMapReceiver(x, *list);
    }
    else {
	UnmapReceiver(x);
    }
    XwcFreeStringList(list);
}

static void
GetPreeditString(IMReceiver *x)
{
    GetString(x, PREEDIT_STRING);
}

static void
GetPreeditFeedback(IMReceiver *x)
{
    Atom atom = PREEDIT_FEEDBACK;
    Atom property_type;
    int format;
    long n_items, remain;
    unsigned long *feedback;

    if (XGetWindowProperty(disp, x->target, atom, 0, 8192,
                           False, AnyPropertyType, &property_type, &format,
                           &n_items, &remain, (void *)&feedback) != Success) {
        return;
    }
    if (property_type != XA_INTEGER || format != 32) {
        XFree(feedback);
        return;
    }
    SetFeedbackPreeditStringPainter(x->psp, feedback, n_items);
    XFree(feedback);
}

static void
GetPreeditCaret(IMReceiver *x)
{
    Atom atom = PREEDIT_CARET;
    Atom property_type;
    int format;
    long n_items, remain;
    unsigned long *caret;

    if (XGetWindowProperty(disp, x->target, atom, 0, 8192,
                           False, AnyPropertyType, &property_type, &format,
                           &n_items, &remain, (void *)&caret) != Success) {
        return;
    }
    if (property_type != XA_INTEGER || format != 32) {
        XFree(caret);
        return;
    }
    SetCaretPositionPreeditStringPainter(x->psp, *caret);
    XFree(caret);
}

static void
GetPreeditSpotLocation(IMReceiver *x)
{
    Atom atom = PREEDIT_SPOT_LOCATION;
    Atom property_type;
    int format;
    long n_items, remain;
    XPoint *spot;

    if (XGetWindowProperty(disp, x->target, atom, 0, 8192,
                           False, AnyPropertyType, &property_type, &format,
                           &n_items, &remain, (void *)&spot) != Success) {
        return;
    }
    if (property_type != XA_POINT || format != 16) {
        XFree(spot);
        return;
    }
    if (n_items > 0) {
	int sx, sy;
	MoveBorderDecorator(x->bd, spot->x, spot->y, &sx, &sy);	
	MovePreeditStringPainter(x->psp, sx, sy);
    }
    XFree(spot);
}

static void
GetPreeditFontset(IMReceiver *x)
{
    XTextProperty text;
    char **list;
    int n_lists, n;
    Atom atom = PREEDIT_FONTSET;

    if (XGetTextProperty(disp, x->target, &text, atom) == 0) {
	SetFontsetPreeditStringPainter(x->psp, NULL);
	return;
    } 
    if ((n = XTextPropertyToStringList(&text, &list, &n_lists)) < 0) {
	XFree(text.value);
	return;
    }
    XFree(text.value);

    SetFontsetPreeditStringPainter(x->psp, (n_lists > 0) ? *list : NULL);
    XFreeStringList(list);
}

static void
GetStatusString(IMReceiver *x)
{
    GetString(x, STATUS_STRING);
}

#if 0
static void
PrintPrediction(char **ptr, int num)
{
    int k;

    printf("\033[2J\033[;H\n");
    if (ptr == NULL)
	return;
    for (k = 0; k < num; ++k) {
	printf("%d: %s\n", k + 1, ptr[k]);
    }
}
#endif

static void
GetPredictString(IMReceiver *x)
{
    XTextProperty text;
    char **list, *mbs, *p;
    int n_lists, n, m, len;
    Atom atom = PREDICT_STRING;

    /*
      2001-05-12:
      XwcTextPropertyToTextList()nĤʸޤƥȥץѥƥ
      ʸΥꥹȤȡԴʥꥹȤˤʤ롣ΤᡢΨ
      ޥХʸǽƤ롣
    */
    if (XGetTextProperty(disp, DefaultRootWindow(disp), &text, atom) == 0) {
	UnmapReceiver(x);
	return;
    }
    if ((n = XmbTextPropertyToTextList(disp, &text, &list, &n_lists)) < 0) {
	UnmapReceiver(x);
	XFree(text.value);
	return;
    }
    XFree(text.value);
    
    len = n_lists - 1;
    for (n = 0; n < n_lists; ++n)
	len += strlen(list[n]);
    if ((mbs = (char *)alloca(sizeof(char *) * (len + 1))) == NULL) {
	UnmapReceiver(x);
	XFreeStringList(list);
	return;
    }
    for (p = mbs, n = 0; n < n_lists; ++n) {
	strcpy(p, list[n]);
	m = strlen(list[n]);
	p[m] = ' ';
	p += m + 1;
    }
    mbs[len] = 0;

    MBSMapReceiver(x, mbs);
    XFreeStringList(list);
}

static void
AddSchedule(IMReceiver *x, int n)
{
    x->schedule |= n;
}

static void
RespondPreeditProperty(IMReceiver *x, XPropertyEvent *p)
{
    if (p->atom == PREEDIT_STRING) {
	AddSchedule(x, GET_STRING);
    }
    else if (p->atom == PREEDIT_FEEDBACK) {
	AddSchedule(x, GET_FEEDBACK);
    }
    else if (p->atom == PREEDIT_CARET) {
	AddSchedule(x, GET_CARET);
    }
    else if (p->atom == PREEDIT_SPOT_LOCATION) {
	AddSchedule(x, GET_SPOT_LOCATION);
    }
    else if (p->atom == PREEDIT_FONTSET) {
	AddSchedule(x, GET_FONTSET);
    }
}

static void
RespondStatusProperty(IMReceiver *x, XPropertyEvent *p)
{
    if (p->atom == STATUS_STRING) {
	GetStatusString(x);
    }
}

static void
RespondPredictProperty(IMReceiver *x, XPropertyEvent *p)
{
    if (p->atom == PREDICT_STRING) {
	GetPredictString(x);
    }
}

static void
ResizeTarget(void *cb_data, int w, int h, int ascent)
{
    ResizeBorderDecorator((BorderDecorator *)cb_data, w, h, ascent);
}

static IMReceiver *
AddTarget(Window w)
{
    XClassHint fq;
    XSetWindowAttributes s;
    IMReceiver *x;

    if (!XFindContext(disp, w, Context, (XPointer *)&x)) {
	return NULL;
    }
    if ((x = (IMReceiver *)malloc(sizeof(IMReceiver))) == NULL) { 
	return NULL;
    }
    x->target = w;
    XSelectInput(disp, x->target, PropertyChangeMask | StructureNotifyMask);
    x->window = XCreateSimpleWindow(disp, DefaultRootWindow(disp),
				    0, 0, 1, 1, 0, 0, 0);
    s.background_pixmap = None;
    s.override_redirect = True;
    s.event_mask = ExposureMask | StructureNotifyMask;
    XChangeWindowAttributes(disp, x->window, CWBackPixmap | CWOverrideRedirect
			    | CWEventMask, &s);
    fq.res_name = THIS_NAME;
    fq.res_class = THIS_CLASS;
    XSetClassHint(disp, x->window, &fq);

    x->bd = CreateBorderDecorator(disp, x->window);
    x->psp = CreatePreeditStringPainter(disp, x->window, "balloon",
					ResizeTarget, (void *)x->bd);
    SetCaretModePreeditStringPainter(x->psp, True);
    x->schedule = ~0;
    x->mapcontrol = True;
    x->next = top;
    XSaveContext(disp, x->target, Context, (XPointer)x);
    top = x;
    return (x);
}

static void
RemoveTarget(IMReceiver *x)
{
    IMReceiver *p;

    XDeleteContext(disp, x->target, Context);
    FreePreeditStringPainter(x->psp);
    FreeBorderDecorator(x->bd);
    XDestroyWindow(disp, x->window);

    if (top == x) {
	top = x->next;
    }
    else {
	for (p = top; p->next != x; p = p->next)
	    ;
	p->next = x->next;
    }
    free(x);
}

static void
SendToplevel(XEvent *ev)
{
    IMReceiver *x;

    switch (ev->type) {
    case SelectionRequest:
	if (ev->xselectionrequest.selection == RECEIVER
	    && ev->xselectionrequest.target == PREEDIT_WINDOW) {
	    (void)AddTarget(ev->xselectionrequest.requestor);
	}	
	break;
    case PropertyNotify:
	if (ev->xany.window == DefaultRootWindow(disp)) {
	    RespondPreeditProperty(&TheRootPreedit, &ev->xproperty);
	    RespondStatusProperty(&TheRootStatus, &ev->xproperty);
	    RespondPredictProperty(&TheRootPredict, &ev->xproperty);
	    return;
	}
	if (XFindContext(disp, ev->xany.window, Context, (XPointer *)&x))
	    return;
	RespondPreeditProperty(x, &ev->xproperty);
	break;
    case SelectionClear:
	if (ev->xselectionclear.selection == RECEIVER) {
	    printf("Another XIM reciever acquired an ownership.\n");
	    MainPower = False;
	}
	break;
    case DestroyNotify:
	if (XFindContext(disp, ev->xany.window, Context, (XPointer *)&x))
	    return;
	RemoveTarget(x);
	break;
    }
}

static void
UpdatePreeditString(IMReceiver *x)
{
    if (x->schedule & GET_STRING)
	GetPreeditString(x);
    if (x->schedule & GET_FEEDBACK)
        GetPreeditFeedback(x);
    if (x->schedule & GET_CARET)
        GetPreeditCaret(x);
    if (x->schedule & GET_SPOT_LOCATION)
        GetPreeditSpotLocation(x);
    if (x->schedule & GET_FONTSET)
        GetPreeditFontset(x);
    x->schedule = 0;
}

static int
NiceTarget(IMReceiver *x)
{
    UpdatePreeditString(x);
    return (NicePreeditStringPainter(x->psp) || NiceBorderDecorator(x->bd));
}

static int
Nice(void)
{
    IMReceiver *x;

    XFlush(disp);
    UpdatePreeditString(&TheRootPreedit);
    if (NicePreeditStringPainter(TheRootPreedit.psp))
	return (True); 
    if (NicePreeditStringPainter(TheRootStatus.psp))
	return (True);
    if (NicePreeditStringPainter(TheRootPredict.psp))
	return (True);
    for (x = top; x != NULL; x = x->next) {
	if (NiceTarget(x)) {
	    return (True);
	}
    }
    return (False);
}

static void
Send(XEvent *ev)
{
    IMReceiver *x;

    SendPreeditStringPainter(TheRootPreedit.psp, ev);
    SendPreeditStringPainter(TheRootStatus.psp, ev);
    SendPreeditStringPainter(TheRootPredict.psp, ev);
    for (x = top; x != NULL; x = x->next) {
	SendPreeditStringPainter(x->psp, ev);
	SendBorderDecorator(x->bd, ev);
    }
}

static void
ResizeTheRoot(void *cb_data, int w, int h, int ascent __unused)
{
    IMReceiver *x = (IMReceiver *)cb_data;

    XResizeWindow(disp, x->window, w, h);
}

static void
SetupTheRootPreedit(int ac, char **av)
{
    XSetWindowAttributes s;
    XSizeHints h;
    IMReceiver *x = &TheRootPreedit;

    x->target = DefaultRootWindow(disp);
    XSelectInput(disp, x->target, PropertyChangeMask);

    x->window = XCreateSimpleWindow(disp, DefaultRootWindow(disp),
				    0, 0, 1, 1, 0, 0, 0);
    s.background_pixmap = None;
    s.event_mask = ExposureMask | StructureNotifyMask;
    XChangeWindowAttributes(disp, x->window, CWBackPixmap | CWEventMask, &s);
    h.flags = 0;
    SetProperties(disp, x->window, "Preedit", THIS_NAME, THIS_CLASS,
		  ac, av, &h, None, None, None);

    x->psp = CreatePreeditStringPainter(disp, x->window, "preedit",
					ResizeTheRoot, x);
    SetCaretModePreeditStringPainter(x->psp, False);
    x->mapcontrol = MapControl;
    GetPreeditString(x);
    GetPreeditFeedback(x);
    GetPreeditCaret(x);
}

static void
SetupTheRootStatus(int ac, char **av)
{
    XSetWindowAttributes s;
    XSizeHints h;
    IMReceiver *x = &TheRootStatus;

    x->target = DefaultRootWindow(disp);
    XSelectInput(disp, x->target, PropertyChangeMask);

    x->window = XCreateSimpleWindow(disp, DefaultRootWindow(disp),
				    0, 0, 1, 1, 0, 0, 0);
    s.background_pixmap = None;
    s.event_mask = ExposureMask | StructureNotifyMask;
    XChangeWindowAttributes(disp, x->window, CWBackPixmap | CWEventMask, &s);
    h.flags = 0;
    SetProperties(disp, x->window, "Status", THIS_NAME, THIS_CLASS,
		  ac, av, &h, None, None, None);

    x->psp = CreatePreeditStringPainter(disp, x->window, "status",
					ResizeTheRoot, x);
    SetCaretModePreeditStringPainter(x->psp, False);
    x->mapcontrol = MapControl;
    GetStatusString(x);
}

static void
SetupTheRootPredict(int ac, char **av)
{
    XSetWindowAttributes s;
    XSizeHints h;
    IMReceiver *x = &TheRootPredict;

    x->target = DefaultRootWindow(disp);
    XSelectInput(disp, x->target, PropertyChangeMask);

    x->window = XCreateSimpleWindow(disp, DefaultRootWindow(disp),
				    0, 0, 1, 1, 0, 0, 0);
    s.background_pixmap = None;
    s.event_mask = ExposureMask | StructureNotifyMask;
    XChangeWindowAttributes(disp, x->window, CWBackPixmap | CWEventMask, &s);
    h.flags = 0;
    SetProperties(disp, x->window, "Predict", THIS_NAME, THIS_CLASS,
		  ac, av, &h, None, None, None);

    x->psp = CreatePreeditStringPainter(disp, x->window, "predict",
					ResizeTheRoot, x);
    SetCaretModePreeditStringPainter(x->psp, False);
    x->mapcontrol = MapControl;
    GetPredictString(x);
}

static void
GetToplevelPreferences(Display *disp)
{
    GetSylSetting(disp, "imreceiver", THIS_CLASS, &AutoHiddenMode);
    if (strcasecmp(AutoHiddenMode.spec, "True") == 0)
	MapControl = True;
}

int
main(int ac, char **av)
{
    XrmDatabase xrdb;
    XEvent ev;
    Window Toplevel;

    if ((disp = XOpenDisplay("")) == NULL) {
        fprintf(stderr, "%s: cannot open display.\n", av[0]);
        exit(1);
    }
    if (setlocale(LC_ALL, "") == NULL) {
        fprintf(stderr, "%s: cannot set locale.\n", av[0]);
        exit(1);
    }
    if (XSupportsLocale() == False) {
        fprintf(stderr, "%s: locale not supported.\n", av[0]);
        exit(1);
    }
    XrmInitialize();
    xrdb = SylMergedResourceDatabase(disp);
    GetToplevelPreferences(disp);

    Context = XUniqueContext();

    RECEIVER = XInternAtom(disp, "XIM_RECEIVER", False);
    PREEDIT_WINDOW = XInternAtom(disp, "XIM_PREEDIT_WINDOW", False);
    PREEDIT_STRING = XInternAtom(disp, "XIM_PREEDIT_STRING", False);
    PREEDIT_FEEDBACK = XInternAtom(disp, "XIM_PREEDIT_FEEDBACK", False);
    PREEDIT_CARET = XInternAtom(disp, "XIM_PREEDIT_CARET", False);
    PREEDIT_SPOT_LOCATION = XInternAtom(disp, "XIM_PREEDIT_SPOT_LOCATION",
					False);
    PREEDIT_FONTSET = XInternAtom(disp, "XIM_PREEDIT_FONTSET", False);
    STATUS_STRING = XInternAtom(disp, "XIM_STATUS_STRING", False);
    PREDICT_STRING = XInternAtom(disp, "XIM_PREDICT_STRING", False);

    Toplevel = XCreateSimpleWindow(disp, DefaultRootWindow(disp),
				   0, 0, 1, 1, 0, 0, 0);
    XSetSelectionOwner(disp, RECEIVER, Toplevel, CurrentTime);

    SetupTheRootPreedit(ac, av);
    SetupTheRootStatus(ac, av);
    SetupTheRootPredict(ac, av);
    while (MainPower) {
#ifdef DEBUG
	printf("heap: %d\n", allocated());
#endif
        while (XEventsQueued(disp, QueuedAfterReading) == 0 && Nice()) {
            ;
        }
	XNextEvent(disp, &ev);
	Send(&ev);
	SendToplevel(&ev);
    }
    XCloseDisplay(disp);
    exit(0);
    return 0;
}
